// cpp2html.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "parser.h"

namespace po = boost::program_options;

//////////////////////////////////////////////////////////////////////////
//! Main function
//////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
	try
	{
		// Parser construction
		Parser_t parser;
		ParserData& data = parser.get< ParserData >();

		// Command line options description
		po::options_description named("Allowed options");
		named.add_options()
			("help", "produce help message")
			("input", po::value< std::string >(), "input file name")
			("output,o", po::value< std::string >(), "output file name")
			("tab-spaces,s", po::value< unsigned int >(&data.m_TabSpaces)->default_value(2), "number of spaces to replace tabs")
			("no-tabs-conversion,t", "don't convert tabs to spaces")
			("no-pp-keywords", "don't recognize preprocessor keywords")
			;

		po::positional_options_description positional;
		positional.add("input", -1);

		//! Command line parameters
		po::variables_map params;
		po::store(po::command_line_parser(argc, argv).
			options(named).positional(positional).run(), params);
		po::notify(params);

		// Check if help was requested
		if (params.count("help"))
		{
			std::cout << named << "\n";
			return 0;
		}

		// Check the command line parameters
		std::string input_file_name;
		if (!params.count("input"))
			throw std::invalid_argument("No input file name specified");
		else
			input_file_name = params["input"].as< std::string >();

		std::string output_file_name;
		if (!params.count("output"))
		{
			std::string::size_type pos = input_file_name.find_last_of(".");
			if (pos == std::string::npos)
				output_file_name = input_file_name + "_html";
			else
				output_file_name.assign(input_file_name.c_str(), pos).append("_html").append(input_file_name.c_str() + pos);
		}
		else
			output_file_name = params["output"].as< std::string >();

		data.m_fNoTabsConversion = (params.count("no-tabs-conversion") != 0);
		data.m_fNoPreprocessorKeyWords = (params.count("no-pp-keywords") != 0);

		// Opening file streams
		std::ifstream input(input_file_name.c_str());
		if (!input.is_open())
			throw std::runtime_error(std::string("Failed to open input file: ") + input_file_name);
		input.exceptions(std::ios_base::badbit);
		std::ofstream output(output_file_name.c_str(), std::ios_base::out | std::ios_base::trunc);
		if (!output.is_open())
			throw std::runtime_error(std::string("Failed to open output file: ") + output_file_name);
		output.exceptions(std::ios_base::badbit | std::ios_base::failbit);
		data.m_pOutput = &output;

		// Initializing the parser
		parser.process(fsm::make_event< Init >());

		// Feeding input file to the parser
		char c;
		while ((c = input.get()) != std::ifstream::traits_type::eof())
		{
			parser.process(c);
		}

		// Finalizing the parser
		parser.process(fsm::make_event< Finalize >());
		parser.reset();
	}
	catch(std::exception& e)
	{
		std::cerr << "error: " << e.what() << "\n";
		return 1;
	}
	catch(...)
	{
		std::cerr << "Exception of unknown type!\n";
	}

	return 0;
}
