#include "stdafx.h"
#include "parser.h"

//////////////////////////////////////////////////////////////////////////
//! A class with common data
//////////////////////////////////////////////////////////////////////////
//! The method puts a character into the stream with HTML transformation, if needed
void ParserData::put_char(char c)
{
	switch (c)
	{
	case '&':
		output() << "&amp;";
		break;
	case '"':
		output() << "&quot;";
		break;
	case '<':
		output() << "&lt;";
		break;
	case '>':
		output() << "&gt;";
		break;
	case '\'':
		output() << "&apos;";
		break;
	default:
		output() << c;
	}
}

//////////////////////////////////////////////////////////////////////////
//! Initial state
//////////////////////////////////////////////////////////////////////////
void Initial::on_process(fsm::event< Init > const&)
{
	// Initializing key words set
	m_KeyWords.insert("asm");
	m_KeyWords.insert("do");
	m_KeyWords.insert("if");
	m_KeyWords.insert("return");
	m_KeyWords.insert("typedef");
	m_KeyWords.insert("auto");
	m_KeyWords.insert("double");
	m_KeyWords.insert("inline");
	m_KeyWords.insert("short");
	m_KeyWords.insert("typeid");
	m_KeyWords.insert("bool");
	m_KeyWords.insert("dynamic_cast");
	m_KeyWords.insert("int");
	m_KeyWords.insert("signed");
	m_KeyWords.insert("typename");
	m_KeyWords.insert("break");
	m_KeyWords.insert("else");
	m_KeyWords.insert("long");
	m_KeyWords.insert("sizeof");
	m_KeyWords.insert("union");
	m_KeyWords.insert("case");
	m_KeyWords.insert("enum");
	m_KeyWords.insert("mutable");
	m_KeyWords.insert("static");
	m_KeyWords.insert("unsigned");
	m_KeyWords.insert("catch");
	m_KeyWords.insert("explicit");
	m_KeyWords.insert("namespace");
	m_KeyWords.insert("static_cast");
	m_KeyWords.insert("using");
	m_KeyWords.insert("char");
	m_KeyWords.insert("export");
	m_KeyWords.insert("new");
	m_KeyWords.insert("struct");
	m_KeyWords.insert("virtual");
	m_KeyWords.insert("class");
	m_KeyWords.insert("extern");
	m_KeyWords.insert("operator");
	m_KeyWords.insert("switch");
	m_KeyWords.insert("void");
	m_KeyWords.insert("const");
	m_KeyWords.insert("false");
	m_KeyWords.insert("private");
	m_KeyWords.insert("template");
	m_KeyWords.insert("volatile");
	m_KeyWords.insert("const_cast");
	m_KeyWords.insert("float");
	m_KeyWords.insert("protected");
	m_KeyWords.insert("this");
	m_KeyWords.insert("wchar_t");
	m_KeyWords.insert("continue");
	m_KeyWords.insert("for");
	m_KeyWords.insert("public");
	m_KeyWords.insert("throw");
	m_KeyWords.insert("while");
	m_KeyWords.insert("default");
	m_KeyWords.insert("friend");
	m_KeyWords.insert("register");
	m_KeyWords.insert("true");
	m_KeyWords.insert("delete");
	m_KeyWords.insert("goto");
	m_KeyWords.insert("reinterpret_cast");
	m_KeyWords.insert("try");

	if (!m_fNoPreprocessorKeyWords)
	{
		// Preprocessor keywords
		m_KeyWords.insert("include");
		m_KeyWords.insert("define");
		m_KeyWords.insert("pragma");
		m_KeyWords.insert("error");
		m_KeyWords.insert("elif");
		m_KeyWords.insert("ifdef");
		m_KeyWords.insert("ifndef");
		m_KeyWords.insert("endif");
		m_KeyWords.insert("defined");
	}

	// Construct a string for tabs replacement
	if (!m_fNoTabsConversion)
		m_TabReplacement.resize(m_TabSpaces, ' ');

	switch_to< ReadingDelimiters >();
}

//////////////////////////////////////////////////////////////////////////
//  ReadingDelimiters state
//////////////////////////////////////////////////////////////////////////
void ReadingDelimiters::Transition::transit(ReadingDelimiters& state, char c)
{
	if (c == '"' || c == '\'')
	{
		state.switch_to< ReadingString >();
	}
	else if (isalpha(c) || c == '_')
	{
		state.switch_to< ReadingIdentifier >();
	}
	else if (!isspace(c))
	{
		state.switch_to< ReadingCommonCode >();
	}
}

void ReadingDelimiters::on_process(char c)
{
	if (c == '\t' && !m_fNoTabsConversion)
		output() << m_TabReplacement;
	else
		output() << c;
}

//////////////////////////////////////////////////////////////////////////
//  ReadingCommonCode state
//////////////////////////////////////////////////////////////////////////
void ReadingCommonCode::Transition::transit(ReadingCommonCode& state, char c)
{
	if (c == '"' || c == '\'')
	{
		state.switch_to< ReadingString >();
	}
	else if (isspace(c))
	{
		state.switch_to< ReadingDelimiters >();
	}
	else if ((isalpha(c) || c == '_') && !isalnum(state.m_LastChar))
	{
		state.switch_to< ReadingIdentifier >();
	}
}

void ReadingCommonCode::on_process(char c)
{
	if (m_fForwardSlashDelayed)
	{
		m_fForwardSlashDelayed = false;
		switch (c)
		{
		case '/':
			// A single-line comment began
			switch_to< ReadingSingleLineComment >();
			break;
		case '*':
			// A multi-line comment began
			switch_to< ReadingMultiLineComment >();
			break;
		default:
			// Anther expression is being read
			put_char('/');
			put_char(c);
			m_LastChar = c;
		}
	}
	else
	{
		if (c == '/')
			// This may be a comment or some kind of expression.
			// Delaying the '/' character to be written until we know what this is.
			m_fForwardSlashDelayed = true;
		else
			put_char(c);

		m_LastChar = c;
	}
}

void ReadingCommonCode::on_leave_state()
{
	// If we have delayed '/', then write it to the stream
	if (m_fForwardSlashDelayed)
	{
		m_fForwardSlashDelayed = false;
		put_char('/');
	}
	m_LastChar = 0;
}

//////////////////////////////////////////////////////////////////////////
//  ReadingIdentifier state
//////////////////////////////////////////////////////////////////////////
void ReadingIdentifier::Transition::transit(ReadingIdentifier& state, char c)
{
	if (c == '"' || c == '\'')
	{
		state.switch_to< ReadingString >();
	}
	else if (isspace(c))
	{
		state.switch_to< ReadingDelimiters >();
	}
	else if (!isalnum(c) && c != '_')
	{
		state.switch_to< ReadingCommonCode >();
	}
}

void ReadingIdentifier::on_process(char c)
{
	m_Identifier.push_back(c);
}

void ReadingIdentifier::on_leave_state()
{
	if (m_KeyWords.count(m_Identifier) > 0)
		output() << "<span class=keyword>" << m_Identifier << "</span>";
	else
		output() << m_Identifier;
	m_Identifier.clear();
}

//////////////////////////////////////////////////////////////////////////
//  ReadingString state
//////////////////////////////////////////////////////////////////////////
void ReadingString::on_process(char c)
{
	if (m_fBackwardSlashReceived)
	{
		m_fBackwardSlashReceived = false;
		put_char(c);
	}
	else
	{
		switch (c)
		{
		case '\\':
			m_fBackwardSlashReceived = true;
			put_char(c);
			break;
		case '\'':
		case '"':
			if (!m_fQuoteReceived)
			{
				m_fQuoteReceived = true;
				m_QuoteChar = c;
				output() << "<span class=string>";
				put_char(c);
			}
			else
			{
				put_char(c);
				if (c == m_QuoteChar)
				{
					output() << "</span>";
					switch_to< ReadingDelimiters >();
				}
			}
			break;
		default:
			put_char(c);
		}
	}
}

void ReadingString::on_leave_state()
{
	m_fBackwardSlashReceived = false;
	m_fQuoteReceived = false;
	m_QuoteChar = '"';
}

//////////////////////////////////////////////////////////////////////////
//  ReadingSingleLineComment state
//////////////////////////////////////////////////////////////////////////
void ReadingSingleLineComment::on_process(char c)
{
	if (c == '\n')
	{
		output() << "</span>";
		switch_to< ReadingDelimiters >();
	}
	put_char(c);
}

void ReadingSingleLineComment::on_enter_state()
{
	// We should write '//' that was skipped in the ReadingCommonCode state
	output() << "<span class=comment>//";
}

//////////////////////////////////////////////////////////////////////////
//  ReadingMultiLineComment state
//////////////////////////////////////////////////////////////////////////
void ReadingMultiLineComment::on_process(char c)
{
	switch (c)
	{
	case '*':
		m_fStarReceived = true;
		output() << c;
		break;

	case '/':
		output() << c;
		if (m_fStarReceived)
		{
			output() << "</span>";
			switch_to< ReadingDelimiters >();
		}
		break;

	case '\t':
		if (!m_fNoTabsConversion)
			output() << m_TabReplacement;
		else
			output() << c;
		break;

	default:
		m_fStarReceived = false;
		put_char(c);
	}
}

void ReadingMultiLineComment::on_enter_state()
{
	// We should write '/*' that was skipped in the ReadingCommonCode state
	output() << "<span class=comment>/*";
}

void ReadingMultiLineComment::on_leave_state()
{
	m_fStarReceived = false;
}
