#ifndef __PARSER_H__
#define __PARSER_H__

//  Event tags
struct Init;
struct Finalize;

//////////////////////////////////////////////////////////////////////////
//! A class with common data
//////////////////////////////////////////////////////////////////////////
struct ParserData
{
public:
	//! A pointer to output the transformed text
	std::ostream* m_pOutput;
	//! A number of spaces to replace tabs
	unsigned int m_TabSpaces;
	//! A flag that shows that no tabs-to-spaces conversion should be done
	bool m_fNoTabsConversion;
	//! A flag that makes the parser not to recognize preprocessor keywords
	bool m_fNoPreprocessorKeyWords;

protected:
	//! A set containing all known C++ keywords
	std::hash_set< std::string > m_KeyWords;
	//! A string m_TabSpaces spaces to replace tabs
	std::string m_TabReplacement;

public:
	//! Constructor
	ParserData() : m_pOutput(NULL), m_fNoTabsConversion(false), m_fNoPreprocessorKeyWords(false) {}

	//! An accessor to output the text
	std::ostream& output() const { return *m_pOutput; }

	//! The method puts a character into the stream with HTML transformation, if needed
	void put_char(char c);
};

//  States
struct Initial;
struct ReadingDelimiters;
struct ReadingCommonCode;
struct ReadingIdentifier;
struct ReadingString;
struct ReadingSingleLineComment;
struct ReadingMultiLineComment;

typedef boost::mpl::vector<
	Initial,
	ReadingDelimiters,
	ReadingCommonCode,
	ReadingIdentifier,
	ReadingString,
	ReadingSingleLineComment,
	ReadingMultiLineComment
>::type StatesList_t;

//////////////////////////////////////////////////////////////////////////
//! A template base class for every state
//////////////////////////////////////////////////////////////////////////
template< typename StateT >
struct StateBase :
	public fsm::state< StateT, StatesList_t >,
	virtual public ParserData
{
	void on_process(fsm::event< Finalize > const&)
	{
	}

	void on_reset()
	{
		static_cast< StateT* >(this)->on_leave_state();
	}
};

//////////////////////////////////////////////////////////////////////////
//! Initial state
//////////////////////////////////////////////////////////////////////////
struct Initial : public StateBase< Initial >
{
	using StateBase< Initial >::on_process;
	void on_process(fsm::event< Init > const&);
};

//////////////////////////////////////////////////////////////////////////
//! ReadingDelimiters state
//////////////////////////////////////////////////////////////////////////
struct ReadingDelimiters : public StateBase< ReadingDelimiters >
{
	struct Transition
	{
		template< typename StateT, typename EventT >
		struct is_applicable : boost::is_same< EventT, char > {};

		static void transit(ReadingDelimiters& state, char c);
	};
	typedef boost::mpl::vector< Transition >::type transitions_type_list;

	using StateBase< ReadingDelimiters >::on_process;
	void on_process(char c);
};

//////////////////////////////////////////////////////////////////////////
//! ReadingCommonCode state
//////////////////////////////////////////////////////////////////////////
struct ReadingCommonCode : public StateBase< ReadingCommonCode >
{
public:
	struct Transition
	{
		template< typename StateT, typename EventT >
		struct is_applicable : boost::is_same< EventT, char > {};

		static void transit(ReadingCommonCode& state, char c);
	};
	typedef boost::mpl::vector< Transition >::type transitions_type_list;

public:
	//! Last received character for better handling some cases with mixed numerics, literals and other non-space chars
	char m_LastChar;
	//! The flag shows if the '/' character was read previously but was not written to the output stream
	bool m_fForwardSlashDelayed;

public:
	ReadingCommonCode() : m_LastChar(0), m_fForwardSlashDelayed(false) {}

	using StateBase< ReadingCommonCode >::on_process;
	void on_process(char c);

	void on_leave_state();
};

//////////////////////////////////////////////////////////////////////////
//! ReadingIdentifier state
//////////////////////////////////////////////////////////////////////////
struct ReadingIdentifier : public StateBase< ReadingIdentifier >
{
public:
	struct Transition
	{
		template< typename StateT, typename EventT >
		struct is_applicable : boost::is_same< EventT, char > {};

		static void transit(ReadingIdentifier& state, char c);
	};
	typedef boost::mpl::vector< Transition >::type transitions_type_list;

private:
	//! The identifier being read
	std::string m_Identifier;

public:
	using StateBase< ReadingIdentifier >::on_process;
	void on_process(char c);

	void on_leave_state();
};

//////////////////////////////////////////////////////////////////////////
//! ReadingString state
//////////////////////////////////////////////////////////////////////////
struct ReadingString : public StateBase< ReadingString >
{
private:
	//! The flag shows if the '\' character was read previously
	bool m_fBackwardSlashReceived;
	//! The flag shows that the '"' or '\'' char has already been received
	bool m_fQuoteReceived;
	//! The quote char type
	char m_QuoteChar;

public:
	ReadingString() : m_fBackwardSlashReceived(false), m_fQuoteReceived(false), m_QuoteChar('"') {}

	using StateBase< ReadingString >::on_process;
	void on_process(char c);

	void on_leave_state();
};

//////////////////////////////////////////////////////////////////////////
//! ReadingSingleLineComment state
//////////////////////////////////////////////////////////////////////////
struct ReadingSingleLineComment : public StateBase< ReadingSingleLineComment >
{
	using StateBase< ReadingSingleLineComment >::on_process;
	void on_process(char c);

	void on_enter_state();
};

//////////////////////////////////////////////////////////////////////////
//! ReadingMultiLineComment state
//////////////////////////////////////////////////////////////////////////
struct ReadingMultiLineComment : public StateBase< ReadingMultiLineComment >
{
private:
	//! The flag shows that the '*' has been received previously
	bool m_fStarReceived;

public:
	ReadingMultiLineComment() : m_fStarReceived(false) {}

	using StateBase< ReadingMultiLineComment >::on_process;
	void on_process(char c);

	void on_enter_state();
	void on_leave_state();
};

//! Text parser type
typedef fsm::state_machine< StatesList_t > Parser_t;

#endif // __PARSER_H__
